"""This module parses impedance data.

Copyright 2021 Michael Hayes, UCECE"""

import numpy as np
from os.path import basename, splitext


class ImpedanceData(object):

    @property
    def Y(self):
        return 1 / self.Z
    

class KeysightE4990AImpedanceData(ImpedanceData):

    def __init__(self, filename):

        lines = open(filename).readlines()
        if not lines[0].startswith('!Agilent Technologies,E4990A'):
            raise ValueError('Not Keysight E4990A')

        if lines[4].startswith('Frequency(Hz), |Z|(Ohm)-data, theta-z(deg)-data'):
            foo = np.loadtxt(filename, skiprows=5, delimiter=',', comments='END')
            self.f = foo[:, 0]
            self.Z = (np.cos(np.radians(foo[:, 2])) + 1j * np.sin(np.radians(foo[:, 2]))) * foo[:, 1]

        elif lines[4].startswith('Frequency(Hz), R(Ohm)-data, X(Ohm)-data'):
        
            foo = np.loadtxt(filename, skiprows=5, delimiter=',', comments='END')
            self.f = foo[:, 0]
            self.Z = foo[:, 1] + 1j * foo[:, 2]
        else:
            raise ValueError('Unhandled format for Keysight E4990A')

        self.filename = basename(filename)
        self.name, ext = splitext(basename(filename.lower()))


class GenericImpedanceData(ImpedanceData):

    def __init__(self, filename):

        foo = np.loadtxt(filename, skiprows=0, delimiter=',', comments='#')
        if foo.shape[1] != 3:
            raise ValueError('Expecting 3 columns (frequency, real, imag')
        
        self.f = foo[:, 0]
        self.Z = foo[:, 1] + 1j * foo[:, 2]
        self.filename = basename(filename)
        self.name, ext = splitext(basename(filename.lower()))        

        
def impedancedata(filename):

    try:
        return KeysightE4990AImpedanceData(filename)
    except:
        pass

    try:
        return GenericImpedanceData(filename)
    except:
        pass    

    # Add support for other data formats.
    
    raise ValueError('Cannot determine impedance format for %s' % filename)





from lcapy import *
import numpy as np

class Foo:

    net = R('R1') + L('L1')

    Z = net.Z(f)

    p = ('R1', 'L1')

    codestr = 'j = 1j; pi = np.pi; ' + str(Z)
    for param in p:
        codestr = codestr.replace(param, 'self.' + param)

    code = compile(codestr, 'foo', 'single')
    

    def __init__(self, R1, L1):

        self.R1 = R1
        self.L1 = L1

    
    def Z(self, f):

        return eval(self.code)
    



#!/usr/bin/python3
"""zfitpy V0.2
Copyright (c) 2021 Michael P. Hayes, UC ECE, NZ

Usage: zfitpy modelname input-filename output-filename
"""

from __future__ import print_function
from matplotlib.pyplot import show, savefig, style
from argparse import ArgumentParser
from zfitpy.model import models, modelmake
from zfitpy import ZFitter
from zfitpy import Plotter
from zfitpy import impedancedata
import sys

def zfitpy_exception(type, value, tb):
   if hasattr(sys, 'ps1') or not sys.stderr.isatty():
      # We are not in interactive mode or we don't have a tty-like
      # device, so call the default hook
      sys.__excepthook__(type, value, tb)
   else:
      import traceback, pdb
      # We are in interactive mode, print the exception...
      traceback.print_exception(type, value, tb)
      print()
      # ...then start the debugger in post-mortem mode.
      pdb.pm()

def model_make(args):

    if not args.net and not args.modelname:
        raise ValueError('Either need to specify model name with --modelname or network with --net')

    if args.net:
        Model = modelmake('Model', args.net)

    if args.modelname:        
        try:
            Model = models[args.modelname]
        except:
            modelnames = ', '.join(list(models.keys()))
            raise ValueError('Unknown model %s: known models: %s' % (args.modelname, modelnames))            

    return Model

def main():

    parser = ArgumentParser(description='Draw schematic of impedance model.')
    parser.add_argument('--version', action='version', version=__doc__.split('\n')[0])
    parser.add_argument('--modelname', type=str, help='model name')
    parser.add_argument('--net', type=str,
                        help="specify network, e.g., R('R1') + L('L1')")
    parser.add_argument('--input_filename', type=str, help='input filename')    
    parser.add_argument('--output_filename', type=str, help='output filename')
    parser.add_argument('--ranges', type=str, help="specify search ranges, e.g.,  {'R1':(0,1),'L1':(10,20)}")
    parser.add_argument('--draw', action='store_true', default=False, help='draw network')
    parser.add_argument('--show', action='store_true', default=False, help='show plot')
    parser.add_argument('--nyquist', action='store_true', default=False, help='use Nyquist plot')    
    parser.add_argument('--plot-error', action='store_true', default=False, help='plot impedance error')
    parser.add_argument('--plot-fit', action='store_true', default=False, help='plot impedance data and fit')
    parser.add_argument('--plot-data', action='store_true', default=False, help='plot impedance data')    
    parser.add_argument('--magphase', action='store_true', default=False, help='plot magnitude and phase of impedance')            
    parser.add_argument('--title', type=str, help='title for plot')
    parser.add_argument('--steps', type=int, default=20,
                        help='the number of search steps per range')
    parser.add_argument('--pdb', action='store_true',
                        default=False,
                        help="enter python debugger on exception")    
    parser.add_argument('--style', type=str, help='matplotlib style filename')

    args = parser.parse_args()

    if args.pdb:
        sys.excepthook = zfitpy_exception
    
    if args.style:
        style.use(args.style)
    
    if args.draw:
        Model = model_make(args)
        model = Model()
        model.draw(args.output_filename)

        if args.show or args.output_filename is None:
            show()        
        return 0

    if args.input_filename is None:
        raise ValueError('Impedance data not specified')
    
    data = impedancedata(args.input_filename)

    if False and args.ranges is None:
        raise ValueError('Search ranges not specified')

    if args.ranges is None:
        fitmodel = None
    else:

        Model = model_make(args)
        zfitter = ZFitter(Model, data.f, data.Z)            
        fitmodel = zfitter(ranges=args.ranges, Ns=args.steps)
        print('%s, error=%.3e' % (fitmodel, fitmodel.error))
    
    plotter = Plotter()
    if args.plot_error and fitmodel:
        plotter.Z_error(data, fitmodel, title=args.title)

    if args.plot_fit:
        if args.nyquist:
            plotter.Z_nyquist(data, fitmodel, title=args.title)
        else:
            plotter.Z_fit(data, fitmodel, title=args.title, magphase=args.magphase)

    if args.plot_data:
        if args.nyquist:
            plotter.Z_nyquist(data, None, title=args.title)
        else:
            plotter.Z_fit(data, None, title=args.title, magphase=args.magphase)

    if args.output_filename is not None:
        savefig(args.output_filename, bbox_inches='tight')

    if args.show or args.output_filename is None:        
        show()
    
    return 0


if __name__ == '__main__':
    main()



from lcapy import f
import numpy as np

def build(net):
    
    Z = net.Z(f)

    paramnames = Z.symbols
    paramnames.pop('f')
    
    Zcodestr = str(Z)
    for p in paramnames:
        Zcodestr = Zcodestr.replace(p, 'self.' + p)
    Zcode = compile(Zcodestr, 'foo', 'eval')

    return net, Zcode


class Bar:

    def draw(self):

        self.net.draw()
    
    def Z(self, f):
        j = 1j
        pi = np.pi
        
        return eval(self.Zcode)
    
    def Y(self, f):

        return 1 / self.Z(f)
    



"""This module provides plotting support for impedance data.

Copyright 2021 Michael Hayes, UCECE"""

from matplotlib.pyplot import subplots
from numpy import degrees, angle

class Plotter(object):

    def Z_error(self, data, model, axes=None, title=None):

        Z = model.Z(data.f)
        
        if axes is None:
            fig, axes = subplots(1)

        axes.plot(data.f, (data.Z.real - Z.real), label=data.name + ' real')
        axes.plot(data.f, (data.Z.imag - Z.imag), '--', label=data.name + ' imag')

        axes.set_xlabel('Frequency (Hz)')
        axes.set_ylabel('Impedance error (ohms)')
        axes.grid(True)

        if title is None:
            title = str(model)
        axes.set_title(title)
        axes.legend()

    def Z_fit(self, data, model=None, axes=None, title=None, magphase=False):

        if model is not None:
            Z = model.Z(data.f)
        else:
            Z = None
            
        if axes is None:
            fig, axes = subplots(1)

        if magphase:
            ax2 = axes.twinx()

            axes.plot(data.f, abs(data.Z), label=data.name + ' data mag')
            ax2.plot(data.f, degrees(angle(data.Z)), '--', label=data.name + ' data phase')
            if Z is not None:
                axes.plot(data.f, abs(Z), label=data.name + ' model mag')
                ax2.plot(data.f, degrees(angle(Z)), '--', label=data.name + ' model phase')                     
            ax2.legend()
            ax2.set_ylabel('Impedance phase (degrees)')            
            
        else:
            axes.plot(data.f, data.Z.real, label=data.name + ' data real')
            axes.plot(data.f, data.Z.imag, '--', label=data.name + ' data imag')

            if Z is not None:            
                axes.plot(data.f, Z.real, label=data.name + ' model real')
                axes.plot(data.f, Z.imag, '--', label=data.name + ' model imag')

        axes.set_xlabel('Frequency (Hz)')
        axes.set_ylabel('Impedance (ohms)')
        axes.grid(True)

        if title is None:
            title = str(model)
        axes.set_title(title)
        axes.legend()

    def Z_nyquist(self, data, model=None, axes=None, title=None):

        if model is not None:
            Z = model.Z(data.f)
        else:
            Z = None
            
        if axes is None:
            fig, axes = subplots(1)

        axes.plot(data.Z.real, data.Z.imag, label=data.name + ' data')
        if Z is not None:
            axes.plot(Z.real, Z.imag, label=data.name + ' model')        

        axes.set_xlabel('Impedance real (ohms)')
        axes.set_ylabel('Impedance imag (ohms)')        
        axes.grid(True)

        if title is None:
            title = str(model)
        axes.set_title(title)
        axes.legend()                

    def Z(self, data, axes=None, title=None):
        
        if axes is None:
            fig, axes = subplots(1)

        axes.plot(data.f, data.Z.real, label=data.name + ' real')
        axes.plot(data.f, data.Z.imag, '--', label=data.name + ' imag')

        axes.set_xlabel('Frequency (Hz)')
        axes.set_ylabel('Impedance (ohms)')
        axes.grid(True)

        if title is None:
            title = data.name
        axes.set_title(title)
        axes.legend()        
        
            



"""This module is a wrapper for the SciPy optimizers

Copyright 2021 Michael Hayes, UCECE"""

import numpy as np
from scipy.signal import resample
from scipy.optimize import fminbound, brute, fmin

class ZFitter(object):

    def __init__(self, model, f, Z, verbose=False):
        self._model = model
        self.verbose = verbose
        self.f = f
        self.Z = Z

    def zmodel_error(self, params):

        model = self._model(*params)
        Z = model.Z(self.f)
        Zerr = Z - self.Z
        rmse = np.sqrt(np.mean(Zerr.real**2 + Zerr.imag**2))
        if self.verbose:        
            print(model, rmse)
        return rmse

    def __call__(self, ranges=None, Ns=10, finish=True):

        if isinstance(ranges, str):
            ranges = eval(ranges)            
        
        if isinstance(ranges, dict):
            rangesdict = ranges
            ranges = []
            for paramname in self._model.paramnames:
                ranges.append(rangesdict[paramname])
        
        if finish:
            # This calls fmin at finish
            params, fval, foo, bar = brute(self.zmodel_error, ranges, Ns=Ns, args=(), full_output=1)
        else:
            params, fval, foo, bar = brute(self.zmodel_error, ranges, Ns=Ns, args=(), full_output=1, finish=None)

        model = self._model(*params)
        model.error = fval
        return model
    
        



"""This module provides the impedance Model class.

Models can be dynamically created given a network described using
Lcapy network syntax, e.g., "R('R1') + L('L1') | R('R2')".

Impedances are calculated using Lcapy.  To speed up evaluations, the
impedance calculating code is lazily compiled and cached.

Copyright 2021 Michael Hayes, UCECE

"""

from lcapy import f, t, G, R, C, L, CPE, Par, Ser
import numpy as np

models = {}

def modelmake(name, net, paramnames=None):

    if isinstance(net, str):
        net = eval(net)
    
    if paramnames is None:
        Z = eval('(%s).Z(f)' % net)
        paramnames = Z.symbols
        paramnames.pop('t', None)
        paramnames.pop('f', None)
    
    params = ', '.join(paramnames)
    
    docstring = name + '(' + params + '): ' + str(net)
    newclass = type(name, (Model, ), {'__doc__': docstring})
    newclass.net = net
    newclass.paramnames = paramnames
    models[name] = newclass
    return newclass


class Model(object):

    _Zcode = None
    error = 0

    def __init__(self, *args):
        """Create model instance."""

        if len(args) == 0:
            return
        
        for m, p in enumerate(self.paramnames):
            setattr(self, p, args[m])
    
    def __repr__(self):
        return self.__str__()

    def __call__(self, i, t):

        return self.v(i, t)

    def v(self, i, t):
        """Calculate voltage drop across the network given a current signal."""

        return self.net.subs(vars(self)).Z.response(i, t)
    
    def i(self, v, t):
        """Calculate current through the network given an applied voltage
        signal."""        

        return self.net.subs(vars(self)).Y.response(v, t)        

    def draw(self, filename=None):
        """Draw the network."""

        self.net.draw(filename)

    def _build(self, foo, name):
    
        paramnames = foo.symbols
        paramnames.pop('t', None)
        paramnames.pop('f', None)        
        
        codestr = str(foo)
        for p in paramnames:
            codestr = codestr.replace(p, 'self.' + p)
        code = compile(codestr, name, 'eval')

        return code

    def _Zbuild(self):    

        return self._build(self.net.Z(f), 'Z')

    def Z(self, f):
        """Return impedance at frequency `f`; `f` can be an ndarray."""

        if self._Zcode is None:
            # Cache result for class
            self.__class__._Zcode = self._Zbuild()
        
        j = 1j
        pi = np.pi
        
        return eval(self._Zcode)
    
    def Y(self, f):
        """Return admittance at frequency `f`; `f` can be an ndarray."""        

        return 1 / self.Z(f)

    def __str__(self):

        parts = []
        for var, val in vars(self).items():
            if var == 'error':
                continue
            
            units = {'R': ' ohms', 'C': ' F', 'L': ' H', 'K':'', 'a':''}[var[0]]

            # Could convert units to have SI prefixes
            parts.append('%s=%.2e%s' % (var, val, units))

        return ', '.join(parts)



"""This module provides a simple function for fitting an electric model to measured impedance data.

Copyright 2021 Michael Hayes, UCECE"""

from .impedancedata import ImpedanceData, impedancedata
from .model import modelmake, Model
from .zfitter import ZFitter

def zfit(data, model, ranges, Ns=10):
    """Fit impedance data to a model.

    `data` can be a filename string or and `ImpedanceData` object
    `model` can be a string describing the network model or a `Model` class
    `ranges` is a dictionary of the search ranges for each component
    `Ns` is the number of search steps in each dimension.

    Returns: `ImpedanceData` object and a best-fit `Model` object.
    """

    if not isinstance(data, ImpedanceData):
        data = impedancedata(data)

    if not isinstance(model, Model):
        model = modelmake('Model', model)

    zfitter = ZFitter(model, data.f, data.Z)
        
    fitmodel = zfitter(ranges=ranges, Ns=Ns)

    return data, fitmodel
